@@ -2,7 +2,7 @@
 
   gc.c -
 
-  $Author: yugui $
+  $Author: naruse $
   created at: Tue Oct  5 09:44:46 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -80,11 +80,21 @@ void *alloca ();
 #define HEAP_MIN_SLOTS 10000
 #define FREE_MIN  4096
 
-static unsigned int initial_malloc_limit   = GC_MALLOC_LIMIT;
+typedef struct {
+    unsigned int initial_malloc_limit;
+    unsigned int initial_heap_min_slots;
+    unsigned int initial_free_min;
+    int gc_stress;
+} ruby_gc_params_t;
+
+ruby_gc_params_t initial_params = {
+    GC_MALLOC_LIMIT,
+    HEAP_MIN_SLOTS,
+    FREE_MIN,
 #if defined(ENABLE_VM_OBJSPACE) && ENABLE_VM_OBJSPACE
-static unsigned int initial_heap_min_slots = HEAP_MIN_SLOTS;
+    FALSE,
 #endif
-static unsigned int initial_free_min       = FREE_MIN;
+};
 
 #define nomem_error GET_VM()->special_exceptions[ruby_error_nomemory]
 
@@ -364,7 +374,7 @@ typedef struct rb_objspace {
 
 #if defined(ENABLE_VM_OBJSPACE) && ENABLE_VM_OBJSPACE
 #define rb_objspace (*GET_VM()->objspace)
-static int ruby_initial_gc_stress = 0;
+#define ruby_initial_gc_stress	initial_params.gc_stress
 int *ruby_initial_gc_stress_ptr = &ruby_initial_gc_stress;
 #else
 static rb_objspace_t rb_objspace = {{GC_MALLOC_LIMIT}, {HEAP_MIN_SLOTS}};
@@ -389,6 +399,9 @@ int *ruby_initial_gc_stress_ptr = &rb_objspace.gc_stress;
 #define mark_stack_overflow	objspace->markstack.overflow
 #define global_List		objspace->global_list
 #define ruby_gc_stress		objspace->gc_stress
+#define initial_malloc_limit	initial_params.initial_malloc_limit
+#define initial_heap_min_slots	initial_params.initial_heap_min_slots
+#define initial_free_min	initial_params.initial_free_min
 
 static void rb_objspace_call_finalizer(rb_objspace_t *objspace);
 
@@ -403,6 +416,7 @@ rb_objspace_alloc(void)
 
     return objspace;
 }
+#endif
 
 static void initial_expand_heap(rb_objspace_t *objspace);
 
@@ -447,15 +461,15 @@ rb_gc_set_params(void)
     }
 }
 
+#if defined(ENABLE_VM_OBJSPACE) && ENABLE_VM_OBJSPACE
 static void gc_sweep(rb_objspace_t *);
 static void slot_sweep(rb_objspace_t *, struct heaps_slot *);
-static void gc_clear_mark_on_sweep_slots(rb_objspace_t *);
+static void rest_sweep(rb_objspace_t *);
 
 void
 rb_objspace_free(rb_objspace_t *objspace)
 {
-    gc_clear_mark_on_sweep_slots(objspace);
-    gc_sweep(objspace);
+    rest_sweep(objspace);
     if (objspace->profile.record) {
 	free(objspace->profile.record);
 	objspace->profile.record = 0;
@@ -479,11 +493,6 @@ rb_objspace_free(rb_objspace_t *objspace)
     }
     free(objspace);
 }
-#else
-void
-rb_gc_set_params(void)
-{
-}
 #endif
 
 /* tiny heap size */
@@ -502,7 +511,7 @@ rb_gc_set_params(void)
 /* 2KB */
 /*#define HEAP_SIZE 0x800 */
 
-#define HEAP_OBJ_LIMIT (HEAP_SIZE / sizeof(struct RVALUE))
+#define HEAP_OBJ_LIMIT (unsigned int)(HEAP_SIZE / sizeof(struct RVALUE))
 
 extern st_table *rb_class_tbl;
 
@@ -1105,7 +1114,6 @@ init_heap(rb_objspace_t *objspace)
     finalizer_table = st_init_numtable();
 }
 
-#if defined(ENABLE_VM_OBJSPACE) && ENABLE_VM_OBJSPACE
 static void
 initial_expand_heap(rb_objspace_t *objspace)
 {
@@ -1115,7 +1123,6 @@ initial_expand_heap(rb_objspace_t *objspace)
         add_heap_slots(objspace, min_size - heaps_used);
     }
 }
-#endif
 
 static void
 set_heaps_increment(rb_objspace_t *objspace)
@@ -2421,28 +2428,6 @@ mark_current_machine_context(rb_objspace_t *objspace, rb_thread_t *th)
 #endif
 }
 
-static void
-gc_clear_mark_on_sweep_slots(rb_objspace_t *objspace)
-{
-    struct heaps_slot *scan;
-    RVALUE *p, *pend;
-
-    if (objspace->heap.sweep_slots) {
-        while (heaps_increment(objspace));
-        while (objspace->heap.sweep_slots) {
-            scan = objspace->heap.sweep_slots;
-            p = scan->slot; pend = p + scan->limit;
-            while (p < pend) {
-                if (p->as.free.flags & FL_MARK && BUILTIN_TYPE(p) != T_ZOMBIE) {
-                    p->as.basic.flags &= ~FL_MARK;
-                }
-                p++;
-            }
-            objspace->heap.sweep_slots = objspace->heap.sweep_slots->next;
-        }
-    }
-}
-
 static void
 gc_marks(rb_objspace_t *objspace)
 {
@@ -2454,8 +2439,6 @@ gc_marks(rb_objspace_t *objspace)
     objspace->count++;
 
 
-    gc_clear_mark_on_sweep_slots(objspace);
-
     SET_STACK_END;
 
     init_mark_stack(objspace);
@@ -2512,6 +2495,8 @@ garbage_collect(rb_objspace_t *objspace)
 
     GC_PROF_TIMER_START;
 
+    rest_sweep(objspace);
+
     during_gc++;
     gc_marks(objspace);
 
@@ -2918,7 +2903,10 @@ run_finalizer(rb_objspace_t *objspace, VALUE objid, VALUE table)
 	VALUE final = RARRAY_PTR(table)[i];
 	args[0] = RARRAY_PTR(final)[1];
 	args[2] = FIX2INT(RARRAY_PTR(final)[0]);
+	status = 0;
 	rb_protect(run_single_final, (VALUE)args, &status);
+	if (status)
+	    rb_set_errinfo(Qnil);
     }
 }
 
@@ -3014,7 +3002,7 @@ rb_objspace_call_finalizer(rb_objspace_t *objspace)
     size_t i;
 
     /* run finalizers */
-    gc_clear_mark_on_sweep_slots(objspace);
+    rest_sweep(objspace);
 
     do {
 	/* XXX: this loop will make no sense */
@@ -3046,7 +3034,8 @@ rb_objspace_call_finalizer(rb_objspace_t *objspace)
 	while (p < pend) {
 	    if (BUILTIN_TYPE(p) == T_DATA &&
 		DATA_PTR(p) && RANY(p)->as.data.dfree &&
-		!rb_obj_is_thread((VALUE)p) && !rb_obj_is_mutex((VALUE)p) ) {
+		!rb_obj_is_thread((VALUE)p) && !rb_obj_is_mutex((VALUE)p) &&
+		!rb_obj_is_fiber((VALUE)p)) {
 		p->as.free.flags = 0;
 		if (RTYPEDDATA_P(p)) {
 		    RDATA(p)->dfree = RANY(p)->as.typeddata.type->function.dfree;
