@@ -25,7 +25,7 @@ module REXML
     DECLARATION = XMLDecl.default
 
     # Constructor
-    # @param source if supplied, must be a Document, String, or IO. 
+    # @param source if supplied, must be a Document, String, or IO.
     # Documents have their context and Element attributes cloned.
     # Strings are expected to be valid XML documents.  IOs are expected
     # to be sources of valid XML documents.
@@ -66,25 +66,27 @@ module REXML
     # of the document
     def add( child )
       if child.kind_of? XMLDecl
-        @children.unshift child
+        if @children[0].kind_of? XMLDecl
+          @children[0] = child
+        else
+          @children.unshift child
+        end
         child.parent = self
       elsif child.kind_of? DocType
-        # Find first Element or DocType node and insert the decl right 
+        # Find first Element or DocType node and insert the decl right
         # before it.  If there is no such node, just insert the child at the
         # end.  If there is a child and it is an DocType, then replace it.
-        insert_before_index = 0
-        @children.find { |x| 
-          insert_before_index += 1
+        insert_before_index = @children.find_index { |x|
           x.kind_of?(Element) || x.kind_of?(DocType)
         }
-        if @children[ insert_before_index ] # Not null = not end of list
-          if @children[ insert_before_index ].kind_of DocType
+        if insert_before_index # Not null = not end of list
+          if @children[ insert_before_index ].kind_of? DocType
             @children[ insert_before_index ] = child
           else
-            @children[ index_before_index-1, 0 ] = child
+            @children[ insert_before_index-1, 0 ] = child
           end
         else  # Insert at end of list
-          @children[insert_before_index] = child
+          @children << child
         end
         child.parent = self
       else
@@ -129,7 +131,8 @@ module REXML
       xml_decl().version
     end
 
-    # @return the XMLDecl encoding of this document as a String.
+    # @return the XMLDecl encoding of this document as an
+    # Encoding object.
     # If no XMLDecl has been set, returns the default encoding.
     def encoding
       xml_decl().encoding
@@ -162,12 +165,12 @@ module REXML
     #   Document.new("<a><b/></a>").serialize( tr )
     #
     # output::
-    #	  output an object which supports '<< string'; this is where the
+    #     output an object which supports '<< string'; this is where the
     #   document will be written.
     # indent::
     #   An integer.  If -1, no indenting will be used; otherwise, the
     #   indentation will be twice this number of spaces, and children will be
-    #   indented an additional amount.  For a value of 3, every item will be 
+    #   indented an additional amount.  For a value of 3, every item will be
     #   indented 3 more levels, or 6 more spaces (2 * 3). Defaults to -1
     # transitive::
     #   If transitive is true and indent is >= 0, then the output will be
@@ -178,14 +181,15 @@ module REXML
     #   Internet Explorer is the worst piece of crap to have ever been
     #   written, with the possible exception of Windows itself.  Since IE is
     #   unable to parse proper XML, we have to provide a hack to generate XML
-    #   that IE's limited abilities can handle.  This hack inserts a space 
+    #   that IE's limited abilities can handle.  This hack inserts a space
     #   before the /> on empty tags.  Defaults to false
     def write( output=$stdout, indent=-1, transitive=false, ie_hack=false )
-      if xml_decl.encoding != "UTF-8" && !output.kind_of?(Output)
+      if xml_decl.encoding != 'UTF-8' && !output.kind_of?(Output)
         output = Output.new( output, xml_decl.encoding )
       end
       formatter = if indent > -1
           if transitive
+            require "rexml/formatters/transitive"
             REXML::Formatters::Transitive.new( indent, ie_hack )
           else
             REXML::Formatters::Pretty.new( indent, ie_hack )
@@ -196,7 +200,7 @@ module REXML
       formatter.write( self, output )
     end
 
-    
+
     def Document::parse_stream( source, listener )
       Parsers::StreamParser.new( source, listener ).parse
     end
@@ -214,7 +218,7 @@ module REXML
     end
 
     attr_reader :entity_expansion_count
-    
+
     def record_entity_expansion
       @entity_expansion_count += 1
       if @entity_expansion_count > @@entity_expansion_limit
