@@ -39,6 +39,20 @@ class TestMethod < Test::Unit::TestCase
     def meth; end
   end
 
+  def mv1() end
+  def mv2() end
+  private :mv2
+  def mv3() end
+  protected :mv3
+
+  class Visibility
+    def mv1() end
+    def mv2() end
+    private :mv2
+    def mv3() end
+    protected :mv3
+  end
+
   def test_arity
     assert_equal(0, method(:m0).arity)
     assert_equal(1, method(:m1).arity)
@@ -109,6 +123,7 @@ class TestMethod < Test::Unit::TestCase
     def o.foo; end
     assert_nothing_raised { RubyVM::InstructionSequence.disasm(o.method(:foo)) }
     assert_nothing_raised { RubyVM::InstructionSequence.disasm("x".method(:upcase)) }
+    assert_nothing_raised { RubyVM::InstructionSequence.disasm(method(:to_s).to_proc) }
   end
 
   def test_new
@@ -209,6 +224,28 @@ class TestMethod < Test::Unit::TestCase
     assert_raise(TypeError) do
       Class.new.class_eval { define_method(:foo, Object.new) }
     end
+
+    assert_raise(TypeError) do
+      Module.new.module_eval {define_method(:foo, Base.instance_method(:foo))}
+    end
+
+    assert_raise(TypeError) do
+      Class.new.class_eval {define_method(:meth, M.instance_method(:meth))}
+    end
+  end
+
+  def test_super_in_proc_from_define_method
+    c1 = Class.new {
+      def m
+        :m1
+      end
+    }
+    c2 = Class.new(c1) { define_method(:m) { Proc.new { super() } } }
+    # c2.new.m.call should return :m1, but currently it raise NoMethodError.
+    # see [Bug #4881] and [Bug #3136]
+    assert_raise(NoMethodError) {
+      c2.new.m.call
+    }
   end
 
   def test_clone
@@ -371,4 +408,56 @@ class TestMethod < Test::Unit::TestCase
     obj.extend(m)
     assert_equal([:m1, :a], obj.public_methods(false), bug)
   end
+
+  def test_visibility
+    assert_equal('method', defined?(mv1))
+    assert_equal('method', defined?(mv2))
+    assert_equal('method', defined?(mv3))
+
+    assert_equal('method', defined?(self.mv1))
+    assert_equal(nil,      defined?(self.mv2))
+    assert_equal('method', defined?(self.mv3))
+
+    assert_equal(true,  respond_to?(:mv1))
+    assert_equal(false, respond_to?(:mv2))
+    assert_equal(true, respond_to?(:mv3))
+
+    assert_equal(true,  respond_to?(:mv1, true))
+    assert_equal(true,  respond_to?(:mv2, true))
+    assert_equal(true,  respond_to?(:mv3, true))
+
+    assert_nothing_raised { mv1 }
+    assert_nothing_raised { mv2 }
+    assert_nothing_raised { mv3 }
+
+    assert_nothing_raised { self.mv1 }
+    assert_raise(NoMethodError) { self.mv2 }
+    assert_nothing_raised { self.mv3 }
+
+    v = Visibility.new
+
+    assert_equal('method', defined?(v.mv1))
+    assert_equal(nil,      defined?(v.mv2))
+    assert_equal(nil,      defined?(v.mv3))
+
+    assert_equal(true,  v.respond_to?(:mv1))
+    assert_equal(false, v.respond_to?(:mv2))
+    assert_equal(true, v.respond_to?(:mv3))
+
+    assert_equal(true,  v.respond_to?(:mv1, true))
+    assert_equal(true,  v.respond_to?(:mv2, true))
+    assert_equal(true,  v.respond_to?(:mv3, true))
+
+    assert_nothing_raised { v.mv1 }
+    assert_raise(NoMethodError) { v.mv2 }
+    assert_raise(NoMethodError) { v.mv3 }
+
+    assert_nothing_raised { v.__send__(:mv1) }
+    assert_nothing_raised { v.__send__(:mv2) }
+    assert_nothing_raised { v.__send__(:mv3) }
+
+    assert_nothing_raised { v.instance_eval { mv1 } }
+    assert_nothing_raised { v.instance_eval { mv2 } }
+    assert_nothing_raised { v.instance_eval { mv3 } }
+  end
 end
