@@ -20,7 +20,7 @@ class TestRequire < Test::Unit::TestCase
   end
 
   def test_require_too_long_filename
-    assert_in_out_err([], <<-INPUT, %w(:ok), [])
+    assert_in_out_err(["RUBYOPT"=>nil], <<-INPUT, %w(:ok), [])
       begin
         require '#{ "foo/" * 10000 }foo'
       rescue LoadError
@@ -29,10 +29,10 @@ class TestRequire < Test::Unit::TestCase
     INPUT
 
     begin
-      assert_in_out_err(["-S", "foo/" * 10000 + "foo"], "") do |r, e|
+      assert_in_out_err(["-S", "-w", "foo/" * 1024 + "foo"], "") do |r, e|
         assert_equal([], r)
         assert_operator(2, :<=, e.size)
-        assert_equal("openpath: pathname too long (ignored)", e.first)
+        assert_match(/warning: openpath: pathname too long \(ignored\)/, e.first)
         assert_match(/\(LoadError\)/, e.last)
       end
     rescue Errno::EINVAL
@@ -40,21 +40,49 @@ class TestRequire < Test::Unit::TestCase
     end
   end
 
-  def test_require_path_home
+  def test_require_nonascii
+    bug3758 = '[ruby-core:31915]'
+    e = assert_raise(LoadError, bug3758) {require "\u{221e}"}
+    assert_match(/\u{221e}\z/, e.message, bug3758)
+  end
+
+  def test_require_path_home_1
     env_rubypath, env_home = ENV["RUBYPATH"], ENV["HOME"]
+    pathname_too_long = /pathname too long \(ignored\).*\(LoadError\)/m
 
     ENV["RUBYPATH"] = "~"
-    ENV["HOME"] = "/foo" * 10000
-    assert_in_out_err(%w(-S test_ruby_test_require), "", [], /^.+$/)
+    ENV["HOME"] = "/foo" * 1024
+    assert_in_out_err(%w(-S -w test_ruby_test_require), "", [], pathname_too_long)
 
-    ENV["RUBYPATH"] = "~" + "/foo" * 10000
+  ensure
+    env_rubypath ? ENV["RUBYPATH"] = env_rubypath : ENV.delete("RUBYPATH")
+    env_home ? ENV["HOME"] = env_home : ENV.delete("HOME")
+  end
+
+  def test_require_path_home_2
+    env_rubypath, env_home = ENV["RUBYPATH"], ENV["HOME"]
+    pathname_too_long = /pathname too long \(ignored\).*\(LoadError\)/m
+
+    ENV["RUBYPATH"] = "~" + "/foo" * 1024
     ENV["HOME"] = "/foo"
-    assert_in_out_err(%w(-S test_ruby_test_require), "", [], /^.+$/)
+    assert_in_out_err(%w(-S -w test_ruby_test_require), "", [], pathname_too_long)
+
+  ensure
+    env_rubypath ? ENV["RUBYPATH"] = env_rubypath : ENV.delete("RUBYPATH")
+    env_home ? ENV["HOME"] = env_home : ENV.delete("HOME")
+  end
+
+  def test_require_path_home_3
+    env_rubypath, env_home = ENV["RUBYPATH"], ENV["HOME"]
 
     t = Tempfile.new(["test_ruby_test_require", ".rb"])
     t.puts "p :ok"
     t.close
+
     ENV["RUBYPATH"] = "~"
+    ENV["HOME"] = t.path
+    assert_in_out_err(%w(-S test_ruby_test_require), "", [], /\(LoadError\)/)
+
     ENV["HOME"], name = File.split(t.path)
     assert_in_out_err(["-S", name], "", %w(:ok), [])
 
@@ -63,6 +91,10 @@ class TestRequire < Test::Unit::TestCase
     env_home ? ENV["HOME"] = env_home : ENV.delete("HOME")
   end
 
+  def test_require_with_unc
+    assert(system(File.expand_path(EnvUtil.rubybin).sub(/\A(\w):/, '//127.0.0.1/\1$/'), "-rabbrev", "-e0"))
+  end if /mswin|mingw/ =~ RUBY_PLATFORM
+
   def test_define_class
     begin
       require "socket"
@@ -300,4 +332,11 @@ class TestRequire < Test::Unit::TestCase
       }
     }
   end
+
+  def test_frozen_loaded_features
+    bug3756 = '[ruby-core:31913]'
+    assert_in_out_err(['-e', '$LOADED_FEATURES.freeze; require "ostruct"'], "",
+                      [], /\$LOADED_FEATURES is frozen; cannot append feature \(RuntimeError\)$/,
+                      bug3756)
+  end
 end
