@@ -120,54 +120,68 @@ class TestSuper < Test::Unit::TestCase
     def uu(a)
       class << self
         define_method(:tt) do |sym|
-          super
+          super(sym)
         end
       end
     end
   end
 
-  class B
-    def m
-      [self, "#{self.class.to_s}::m"]
-    end
-  end
-
-  class C < B
-    def self.t
-      define_method(:m) {super}
-    end
-  end
-
   def test_define_method
     a = A.new
     a.uu(12)
-    assert_equal("A#tt", a.tt(12), '[ruby-core:03856]')
-
-    bug2419 = '[ruby-core:26984]'
-    q = C.t
-    assert_raise(NoMethodError, bug2419) {q.call}
-    c = C.new
-    assert_equal([c, "#{C.to_s}::m"], c.m, bug2419)
+    assert_equal("A#tt", a.tt(12), "[ruby-core:3856]")
+    e = assert_raise(RuntimeError, "[ruby-core:24244]") {
+      lambda {
+        Class.new do
+          define_method(:a) {super}.call
+        end
+      }.call
+    }
+    assert_match(/implicit argument passing of super from method defined by define_method/, e.message)
   end
 
-  module Bug2537
-    class Parent
-      def run(a)
-        a
-      end
+  class SubSeq
+    def initialize
+      @first=11
+      @first or fail
     end
 
-    class Child < Parent
-      def run(*a)
-        proc {super(*a)}.call
+    def subseq
+      @first or fail
+    end
+  end
+
+  class Indexed
+    def subseq
+      SubSeq.new
+    end
+  end
+
+  Overlaid = proc do
+    class << self
+      def subseq
+        super.instance_eval(& Overlaid)
       end
     end
   end
 
-  def test_super_in_block_call
-    bug2537 = '[ruby-dev:39931]'
-    assert_nothing_raised(bug2537) do
-      assert_equal(bug2537, Bug2537::Child.new.run(bug2537), bug2537)
+  def test_overlaid
+    assert_nothing_raised('[ruby-dev:40959]') do
+      overlaid = proc do |obj|
+        def obj.reverse
+          super
+        end
+      end
+      overlaid.call(str = "123")
+      overlaid.call(ary = [1,2,3])
+      str.reverse
+    end
+
+    assert_nothing_raised('[ruby-core:27230]') do
+      mid=Indexed.new
+      mid.instance_eval(&Overlaid)
+      mid.subseq
+      mid.subseq
     end
   end
 end
